# Copyright 2021 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from argparse import ArgumentTypeError
import logging
from typing import TYPE_CHECKING

from dateutil import parser
from dateutil import tz
from keystoneauth1.exceptions import HttpError
from osc_lib.command import command
from osc_lib import utils

from doniclient.osc.cli import HardwarePatchCommand, BaseParser

if TYPE_CHECKING:
    from doniclient.v1.client import Client as DoniClient

LOG = logging.getLogger(__name__)
COLUMNS = ("uuid", "start", "end")


def flexible_datetime(date_str):
    try:
        parsed_dt = parser.parse(date_str)
        dt_with_tz = parsed_dt.replace(tzinfo=parsed_dt.tzinfo or tz.gettz())
        return dt_with_tz
    except ValueError:
        raise ArgumentTypeError(f"Not a valid date: '{date_str}'.")


def _add_date_args(parser, required=True):
    parser.add_argument(
        "--start",
        type=flexible_datetime,
        help=("Date for start of availability window"),
        required=required,
    )
    parser.add_argument(
        "--end",
        type=flexible_datetime,
        help=("Date for end of availability window"),
        required=required,
    )


class ListHardwareAvailability(command.Lister):
    """List all availability windows for a given hardware item."""

    columns = COLUMNS

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            dest="hardware_uuid", metavar="<hardware_uuid>", help=("unique ID of hardware item")
        )
        return parser

    def take_action(self, parsed_args):
        """List all hw items in Doni."""
        hw_client: "DoniClient" = self.app.client_manager.inventory
        try:
            items = hw_client.get_availability(parsed_args.hardware_uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        items_iterator = (
            utils.get_dict_properties(s, self.columns, formatters={}) for s in items
        )
        return (self.columns, items_iterator)



class AddHardwareAvailability(HardwarePatchCommand):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        _add_date_args(parser)
        return parser

    def get_patch(self, parsed_args):
        return [
            {
                "op": "add",
                "path": "/availability/-",
                "value": {
                    "start": parsed_args.start,
                    "end": parsed_args.end,
                }
            }
        ]


class UpdateHardwareAvailability(HardwarePatchCommand):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            dest="window_uuid",
            metavar="<window_uuid>",
            help=("unique ID of availability window"),
        )
        _add_date_args(parser, required=False)
        return parser

    def get_patch(self, parsed_args):
        patch = []

        for field in ["start", "end"]:
            value = getattr(parsed_args, field)
            if value:
                patch.append({
                    "op": "replace",
                    "path": f"/availability/{parsed_args.window_uuid}/{field}",
                    "value": value
                })

        return patch


class RemoveHardwareAvailability(HardwarePatchCommand):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            dest="window_uuid",
            metavar="<window_uuid>",
            help=("unique ID of availability window"),
        )
        return parser

    def get_patch(self, parsed_args):
        return [
            {
                "op": "remove",
                "path": f"/availability/{parsed_args.window_uuid}",
            },
        ]



"""Implements Doni command line interface."""

import json
import logging
from argparse import ArgumentTypeError, FileType
from sys import stdin
from typing import DefaultDict, List

from dateutil import parser, tz
from keystoneauth1.exceptions import BadRequest, HttpError, NotFound
from osc_lib import utils
from osc_lib.cli import parseractions
from osc_lib.command import command

LOG = logging.getLogger(__name__)  # Get the logger of this module


class DoniClientError(BaseException):
    """Base Error Class for Doni Client."""


class OutputFormat:
    columns = (
        "name",
        "project_id",
        "hardware_type",
        "properties",
        "uuid",
    )


class BaseParser(command.Command):

    require_hardware = False

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("-d", "--dry-run", "--dry_run", action="store_true")
        if self.require_hardware:
            parser.add_argument(
                dest="uuid", metavar="<uuid>", help=("unique ID of hardware item")
            )
        return parser

class ListHardware(command.Lister):
    """List all hardware in the Doni database."""

    columns = OutputFormat.columns

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--all",
            help="List hardware from all owners. Requires admin rights.",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        """List all hw items in Doni."""
        hw_client = self.app.client_manager.inventory
        try:
            if parsed_args.all:
                data = hw_client.list()
            else:
                data = hw_client.export()
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        data_iterator = (
            utils.get_dict_properties(s, self.columns, formatters={}) for s in data
        )
        return (self.columns, data_iterator)


class GetHardware(BaseParser, command.ShowOne):
    """List specific hardware item in Doni."""
    require_hardware = True
    columns = OutputFormat.columns

    def take_action(self, parsed_args):
        """List all hw items in Doni."""
        hw_client = self.app.client_manager.inventory
        try:
            data = hw_client.get_by_uuid(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        return (
            self.columns,
            utils.get_dict_properties(data, self.columns, formatters={}),
        )


class DeleteHardware(BaseParser):
    """Delete specific hardware item in Doni."""

    require_hardware = True

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        try:
            result = hw_client.delete(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        return result.text


class SyncHardware(BaseParser):
    """Sync specific hardware item in Doni."""

    require_hardware = True

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        try:
            result = hw_client.sync(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        return result.text


class CreateHardware(BaseParser):
    """Create a Hardware Object in Doni."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=(
                "Name of the hardware object. Best practice is to use a "
                "universally unique identifier, such has serial number or chassis ID. "
                "This will aid in disambiguating systems."
            ),
            required=True,
        )
        parser.add_argument(
            "--hardware_type",
            metavar="<hardware_type>",
            help=("hardware_type of item"),
            default="baremetal",
        )
        parser.add_argument("--mgmt_addr", metavar="<mgmt_addr>", required=True)
        parser.add_argument("--ipmi_username", metavar="<ipmi_username>")
        parser.add_argument("--ipmi_password", metavar="<ipmi_password>")
        parser.add_argument(
            "--ipmi_terminal_port", metavar="<ipmi_terminal_port>", type=int
        )
        parser.add_argument(
            "--interface",
            required_keys=["name", "mac"],
            action=parseractions.MultiKeyValueAction,
            help=(
                "Specify once per interface, in the form:\n `--interface name=<name>,mac=<mac_address>`"
            ),
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        """Create new HW item."""
        hw_client = self.app.client_manager.inventory

        body = {
            "name": parsed_args.name,
            "hardware_type": parsed_args.hardware_type,
            "properties": {},
        }
        body["properties"]["management_address"] = parsed_args.mgmt_addr

        for arg in ["ipmi_username", "ipmi_password", "ipmi_terminal_port"]:
            value = getattr(parsed_args, arg)
            if value:
                body["properties"][arg] = value
        body["properties"]["interfaces"] = self._format_iface(parsed_args.interface)

        if parsed_args.dry_run:
            LOG.warn(parsed_args)
            LOG.warn(body)
        else:
            try:
                data = hw_client.create(body)
            except HttpError as ex:
                LOG.error(ex.response.text)
                raise ex

            return data

    def _format_iface(self, interface_args: List):
        interface_list = []
        for interface in interface_args or []:
            interface_list.append(
                {
                    "name": interface.get("name"),
                    "mac_address": interface.get("mac"),
                }
            )
        return interface_list


class HardwarePatchCommand(BaseParser):
    require_hardware = True

    def get_patch(self, parsed_args):
        return []

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        hw_uuid = parsed_args.uuid

        patch = self.get_patch(parsed_args)

        if parsed_args.dry_run:
            LOG.info(patch)
            return None

        if patch:
            try:
                LOG.debug(f"PATCH_BODY:{patch}")
                res = hw_client.update(hw_uuid, patch)
            except HttpError as ex:
                LOG.error(ex.response.text)
                raise ex
            else:
                return res.json()
        else:
            LOG.info("No updates to send")


class UpdateHardware(HardwarePatchCommand):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            "--name",
            metavar="<name>",
            help=(
                "Name of the hardware object. Best practice is to use a "
                "universally unique identifier, such has serial number or chassis ID. "
                "This will aid in disambiguating systems."
            ),
        )
        parser.add_argument(
            "--hardware_type",
            metavar="<hardware_type>",
            help=("hardware_type of item"),
        )
        parser.add_argument("--mgmt_addr", metavar="<mgmt_addr>")
        parser.add_argument("--ipmi_username", metavar="<ipmi_username>")
        parser.add_argument("--ipmi_password", metavar="<ipmi_password>")
        parser.add_argument(
            "--ipmi_terminal_port", metavar="<ipmi_terminal_port>", type=int
        )

        subparsers = parser.add_subparsers(help="Select property to update.")

        parse_iface = subparsers.add_parser("interface")
        parse_iface.set_defaults(func=self._handle_ifaces)
        parse_iface.add_argument(
            "--add",
            required_keys=["name", "mac"],
            action=parseractions.MultiKeyValueAction,
            help=(
                "Specify once per interface, in the form:\n `--add name=<name>,mac=<mac_address>`"
            ),
        )
        parse_iface.add_argument(
            "--update",
            required_keys=["name", "mac", "index"],
            action=parseractions.MultiKeyValueAction,
            help=(
                "Specify once per interface, in the form:\n `--update name=<name>,mac=<mac_address>,index=<index>`"
            ),
        )
        parse_iface.add_argument(
            "--delete",
            metavar="index",
            help=("Specify interface to delete, by index`"),
        )

        return parser

    def _handle_ifaces(self, parsed_args):
        # Update Interfaces
        patch = []
        for iface in getattr(parsed_args, "add") or []:
            patch.append({"op": "add", "path": f"/interface/-", "value": iface})

        for iface in getattr(parsed_args, "update") or []:
            index = iface.pop("index")
            patch.append(
                {"op": "replace", "path": f"/interface/{index}", "value": iface}
            )
        for iface in getattr(parsed_args, "delete") or []:
            index = iface.get("index")
            patch.append({"op": "remove", "path": f"/interface/{index}"})
        return patch

    def get_patch(self, parsed_args):
        patch = []
        field_map = {
            "name": "name",
            "hardware_type": "hardware_type",
            "management_address": "properties/management_address",
            "ipmi_username": "properties/ipmi_username",
            "ipmi_password": "properties/ipmi_password",
            "ipmi_terminal_port": "properties/ipmi_terminal_port",
        }
        for key, val in field_map.items():
            arg = getattr(parsed_args, key, None)
            if arg:
                patch.append({"op": "add", "path": f"/{val}", "value": arg})

        subparser = getattr(parsed_args, "func", None)
        if subparser:
            patch.extend(subparser(parsed_args))

        return patch

class ImportHardware(BaseParser):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("-f", "--file", help="JSON input file", type=FileType("r"))
        return parser

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        with parsed_args.file as f:
            for item in json.load(f):
                if parsed_args.dry_run:
                    LOG.warn(item)
                else:
                    try:
                        data = hw_client.create(item)
                    except HttpError as ex:
                        LOG.error(ex.response.text)
                        raise ex
                    else:
                        LOG.debug(data)



"""Creates doni client object."""
import json
import logging

from keystoneauth1.adapter import Adapter as ksa_adapter

LOG = logging.getLogger(__name__)  # Get the logger of this module


class Client(object):
    def __init__(self, adapter: ksa_adapter, **kwargs):
        self.adapter = adapter

    def list(self):
        resp = self.adapter.get("/v1/hardware/")
        try:
            return resp.json().get("hardware")
        except json.JSONDecodeError:
            return resp

    def export(self):
        resp = self.adapter.get("/v1/hardware/export/")
        try:
            return resp.json().get("hardware")
        except json.JSONDecodeError:
            return resp

    def get_by_uuid(self, uuid):
        resp = self.adapter.get(f"/v1/hardware/{uuid}/")
        try:
            return resp.json()
        except json.JSONDecodeError:
            return resp

    def get_availability(self, hardware_uuid: str):
        resp = self.adapter.get(f"/v1/hardware/{hardware_uuid}/availability")
        try:
            return resp.json().get("availability", [])
        except json.JSONDecodeError:
            return resp

    def create(self, json, **kwargs):
        """Create a hw object in the doni DB."""
        resp = self.adapter.post("/v1/hardware/", json=json)
        try:
            return resp.json()
        except json.JSONDecodeError:
            return resp

    def delete(self, uuid):
        return self.adapter.delete(f"/v1/hardware/{uuid}/")

    def sync(self, uuid):
        return self.adapter.post(f"/v1/hardware/{uuid}/sync")

    def update(self, uuid, json):
        return self.adapter.patch(f"/v1/hardware/{uuid}/", json=json)
